# -*- coding: utf-8 -*-
"""
Simple example with the solvers pgd-P0, pgd-P2 and vanilla.
See: https://github.com/cvanwynsberghe/sgcal-jasa
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import pdist, squareform
from sgcal_ambient import CalibrateDiffuse

M = 50          # number of sensors
wavelength = 2  # wavelength [m]
var_n = 1e-2    # noise variance

# %% Generate data
np.random.seed(212)

pos = 1*np.random.rand(M, 2)  # sensor positions in the plane
a_gt = np.random.rand(M) + 0.5 + 1j*np.random.rand(M) - 0.5j  # sensor gains

# model of ambient noise covariance
dist = squareform(pdist(pos))
S = np.sinc(2*np.pi/wavelength*dist)

# create measured covariance for an infinite number of snapshots
N_ambient = np.diag(a_gt) @ S @ np.diag(a_gt.conj())
N_sensor = var_n * np.eye(M)
R = N_ambient + N_sensor

print( 'size(S)=', np.shape(S) )
print( 'size(N_ambient)=', np.shape(N_ambient) )
print( 'size(R)=', np.shape(R) )

# %% calibration solvers
n_irer = 10000
cd_solver0 = CalibrateDiffuse(R, S)  # pgd-P0
cd_solver0.pgd_l0(n_it=n_irer)

cd_solver1 = CalibrateDiffuse(R, S)  # pgd-P1
cd_solver1.pgd_l1(reg_l1=0.01, n_it=n_irer)

cd_solver2 = CalibrateDiffuse(R, S)  # vanilla
cd_solver2.vanilla()

# scale the solutions
cd_solver0.scale_to(a_gt[:, None])
cd_solver1.scale_to(a_gt[:, None])
cd_solver2.scale_to(a_gt[:, None])

print("SNR:", 10*np.log10(np.linalg.norm(N_ambient)/np.linalg.norm(N_sensor)))
print("rmse vanilla:", np.linalg.norm(a_gt - cd_solver2.a_est.T))
print("rmse pgd-P0:", np.linalg.norm(a_gt - cd_solver0.a_est.T))
print("rmse pgd-P1:", np.linalg.norm(a_gt - cd_solver1.a_est.T))

# %% Plot results
fig, (ax0, ax1) = plt.subplots(1, 2, figsize=(7, 4))

for ax_, title_, solver_ in zip([ax0, ax1], ["pgd-P0", "pgd-P1"],
                                [cd_solver0, cd_solver1]):

    ax_.scatter(a_gt.real, a_gt.imag, label="ground truth", s=5, c="black")
    ax_.scatter(cd_solver2.a_est.real, cd_solver2.a_est.imag, marker="s",
                facecolors="none", edgecolors="orange", label="svd")
    ax_.scatter(solver_.a_est.real, solver_.a_est.imag,
                facecolors="none", edgecolors="green",
                label=title_)

    ax_.grid()
    ax_.axis("scaled")
    ax_.set_xlim(-1.8, 1.8)
    ax_.set_ylim(-1.8, 1.8)
    ax_.set_title(title_)
    ax_.legend()

plt.suptitle("Gain/phase calibration of sensor array from ambient noise by\n"
             "covariance measurement fitting - simple example.")
plt.tight_layout()

plt.show()



# -*- coding: utf-8 -*-

# Copyright (c) 2023 Charles Vanwynsberghe
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the “Software”), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# See https://github.com/cvanwynsberghe/sgcal-jasa

import numpy as np

class CalibrateDiffuse:
    def __init__(self, Y, S):
        """
        Sensor gain and phase calibration in ambient noise with covariance
        readings, with 3 implemented solvers: vanilla, pgd-l1 and pgd-l0.
        Parameters
        ----------
        Y : measured covariance matrix
        S : modeled covariance matrix in the diffuse field
        """
        self.Y = Y
        self.S = S
        self.M = self.Y.shape[0]
        self.step_opt = 2 / np.max(np.abs(self.S)**2)  # max stable step

    def _get_a_est(self, normed=False):
        """
        Get final gain vector (principal eigenvector of C).
        """
        self.a_est, s, _ = np.linalg.svd((self.C_ + self.C_.T.conj())/2)
        self.a_est = self.a_est[:, 0]
        self.a_est *= s[0]**0.5
        self.a_est = self.a_est[:, None]
        if normed:
            self.a_est /= self.a_est.mean()

    def pgd_l1(self, reg_l1=0.5, step="opt", n_it=10000, a_init=None,
               normed=False, save_iters=False, verbose=True):
        """
        Proximal gradient descent solver with trace penalty on C.
        Parameters
        ----------
        reg_l1 : regularization parameter
        step : gradient step
        n_it : max number of iterations
        a_init : gains at initialization
        normed : indicate if estimated gain vector should be normalized
        save_iters : keep all iterations
        verbose : helper message
        """
        if step == "opt":
            step = self.step_opt

        if a_init is None:
            self.a_init = np.ones((self.M, 1), np.complex)
        else:
            self.a_init = a_init

        if save_iters is True:
            self.a_list = []

        self.C_ = self.a_init @ self.a_init.T.conj()
        self.rk_C_ = np.zeros((n_it))
        for n_ in range(n_it):
            C_old = self.C_
            self.Z_ = self.C_ + step/2*self.S.conj()*(self.Y - self.S*self.C_)
            self.C_, self.rk_C_[n_] = prox_st((self.Z_ + self.Z_.T.conj())/2,
                                                 step*reg_l1)

            self._get_a_est(normed=normed)
            if save_iters is True:
                self.a_list.append(self.a_est)

            if np.linalg.norm(C_old - self.C_) < 1e-6:
                self.rk_C_ = self.rk_C_[0:n_+1]
                if save_iters is True:
                    self.a_list = np.array(self.a_list).squeeze()
                if verbose:
                    print(f"pgd-l1 done, {n_} iterations, "
                          f"rank(C) = {self.rk_C_[-1]}")
                break

        self._get_a_est(normed=normed)

    def pgd_l0(self, reg_l0=1, step="opt", n_it=10000, a_init=None,
               normed=False, save_iters=False, verbose=True):
        """
        Proximal gradient descent solver with rank penalty on C.
        Parameters
        ----------
        reg_l1 : rank of the matrix C
        step : gradient step
        n_it : max number of iterations
        a_init : gains at initialization
        normed : indicate if estimated gain vector should be normalized
        save_iters : keep all iterations
        verbose : helper message
        Parameters
        ----------
        reg_l0 : rank of the projected matrix C_
        step : gradient step
        n_it : max number of iterations
        a_init : gains at initialization
        normed : indicate if estimated gain vector should be normalized
        save_iters : keep all iterations
        verbose : helper message
        """
        if step == "opt":
            step = self.step_opt
        if a_init is None:
            self.a_init = np.ones((self.M, 1))
        else:
            self.a_init = a_init
        if save_iters is True:
            self.a_list = []

        self.C_ = self.a_init @ self.a_init.T.conj()
        self.rk_C_ = np.zeros((n_it))
        for n_ in range(n_it):
            C_old = self.C_
            self.Z_ = self.C_ + step/2*self.S*(self.Y - self.S*self.C_)
            self.C_, self.rk_C_[n_] = prox_ht((self.Z_ + self.Z_.T.conj())/2,
                                                   reg_l0)
            self._get_a_est(normed=normed)
            if save_iters is True:
                self.a_list.append(self.a_est)

            if np.linalg.norm(C_old - self.C_) < 1e-7:
                self.rk_C_ = self.rk_C_[0:n_+1]
                if verbose:
                    print(f"pgd-l0 done, {n_} iterations")
                break

        self._get_a_est(normed=normed)

    def vanilla(self, normed=False):
        """
        Vanilla apprach by element-wise division.
        Parameters
        ----------
        normed : indicate if estimated gain vector should be normalized
        """
        self.C_ = self.Y/self.S
        self._get_a_est(normed=normed)

    def scale_to(self, a_ref):
        """
        Align estimated gains onto a_ref in the least square sense.
        """
        alpha = (self.a_est.T.conj()@a_ref) / (self.a_est.T.conj()@self.a_est)
        self.a_est *= alpha
        try:
            for i, a_est in enumerate(self.a_tune):
                alpha = (a_est.T.conj()@a_ref) / (a_est.T.conj()@a_est)
                self.a_tune[i, :] = a_est * alpha
        except AttributeError:
            pass


def prox_st(mat, th):
    """
    Soft-thresholding operator on singular values of mat with threshold th.
    """
    u, s, vh = np.linalg.svd(mat)
    s = np.maximum(s - th, 0.0)
    s_card = (s != 0).sum()
    return u @ np.diag(s) @ vh, s_card


def prox_ht(mat, k):
    """
    Hard-thresholding operator on singular values of mat preserving the k
    largest non-zero values.
    """
    u, s, vh = np.linalg.svd(mat)
    s[k::] = 0
    return u @ np.diag(s) @ vh, k


# megamicros.data.py python software for data classes
#
# Copyright (c) 2023-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
This module provides data classes for handling audio and multimedia data.

Features:
    - MuData: Base class for Megamicros data objects
    - MuAudio: Class for multi-channel audio data handling

Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""

import os
from matplotlib import pyplot as plt
import numpy as np
from scipy.io import wavfile
from imageio import mimsave

from megamicros.exception import MuException
from megamicros.log import log

DEFAULT_LIMIT_VALUE = 10
FILETYPE_H5 = 1
FILETYPE_MP4 = 2
FILETYPE_WAV = 3
FILETYPE_MUH5 = 4


# =============================================================================
# Data and Audio Data base classes
# =============================================================================

class MuData:
    """Base class for Megamicros data objects"""

# =============================================================================
# The MuAudio classe
# =============================================================================

class MuAudio( MuData ):
    """
    MuAudio data class for multi-channels audio objects
    """
    __label: str = ""
    __raw: np.array = np.array( [] )
    __sampling_frequency: int = 0
    __dtype: np.dtype = np.float32
    __frame_size: int = None
    __frame_number: int = 0
    __it: int = 0


    def __init__(
            self, audio_signal:np.ndarray|list|None=None,
            sampling_frequency: int|float=None,
            label:str="No label",
            frame_size = None,
            frame_number = 0
        ):
        if audio_signal is not None and sampling_frequency is None:
            raise MuException( "Cannot create MuAudio object without knowing its sampling frequency" )

        if audio_signal is None:
            self.__raw = np.array( [] )
            self.__sampling_frequency = 0
            self.__label = label
            frame_size = None
            frame_number = 0

        # Get signal as an array of only one signal
        elif type( audio_signal ) is list:
            self.__raw = np.array( [audio_signal] )
            self.__sampling_frequency = sampling_frequency
            self.__label = label
            if frame_size is None:
                self.__frame_size = self.__raw.size
                self.__frame_number = 1

            print( f"I'm a list of one signal with frame size = {self.__frame_size} and frame number = {self.__frame_number}")
        
        # Get signal as a ND array (signals number x signals size)
        else:
            self.__raw = audio_signal
            self.__sampling_frequency = sampling_frequency
            self.__label = label
            # set default frame size as the length of signals  
            if frame_size is None:
                self.__frame_size = np.shape(self.__raw)[1]
                self.__frame_number = 1
            else:
                self.__frame_number = int( np.shape(self.__raw)[1] / self.__frame_size )

            print( f"I'm a NDarray signal with frame size = {self.__frame_size} and frame number = {self.__frame_number}")


# =============================================================================
# Properties
# =============================================================================

    @property
    def label( self ):
        """The data label if any"""
        return self.__label

    @property
    def channels_number( self ):
        """The number of channels"""
        if len( np.shape( self.__raw ) ) == 1:
            return 1
        return np.shape( self.__raw )[0]

    @property
    def samples_number( self ):
        """The number of samples per channels"""
        if len( np.shape( self.__raw ) ) == 1:
            return len( self.__raw )
        else:
            return np.shape( self.__raw )[1]

    @property
    def sampling_frequency( self ):
        """Audio signal's sampling frequency."""
        return self.__sampling_frequency 

    @property
    def dtype( self ):
        """The numpy data type used to store audio signals"""
        return self.__dtype

# =============================================================================
# Iterator and bracket operator
# =============================================================================

    def __iter__( self ):
        self.__it = 0
        if self.__frame_number == 0:
            raise Exception( f"Cannot iterate: empty object with no frame to iterate on" ) 
        return self

    def __next__( self ) -> np.ndarray :
        if self.__it >= self.__frame_number:
            raise StopIteration
        
        offset = self.__it * self.__frame_size
        result = self.__raw[:,offset:offset+self.__frame_size]
        self.__it += 1
        return result

    def __getitem__( self, item: int ) -> np.ndarray :
        if item == -1:
            offset = self.__frame_number-1
        elif item < -1 or item >= self.__frame_number:
            raise IndexError( f"Index value ({item}) exceed the avalaible frames number (allowed values are between 0 and {self.__frame_number-1}) " )
        else:
            offset = item * self.__frame_size
        
        return self.__raw[:,offset:offset+self.__frame_size]
    
    def __call__( self ) -> np.ndarray :
        return self.__raw


# =============================================================================
# Overloaded interface
# =============================================================================

    def __str__( self ):
        return f"{self.channels_number} X {self.samples_number} audio signals (sf={self.sampling_frequency} Hz)"

# =============================================================================
# Interface
# =============================================================================

    def channel( self, channel_number:int=0 ) -> np.array:
        """
        Get the audio signal which channel is given as input (np.array)

        ## Parameters
        * channel_number: the channel number
        """
        if channel_number >= self.channels_number:
            raise MuException( f"Array overflow: there is no channel {channel_number} in MuAUdio object" )

        return self.__raw[channel_number,:]

    def set_frame_size( self, frame_size: int ) -> None:
        """
        Set the frame size for cutting signal into frames of fixed length when iterating 
        The default frame size (if not set by user) is equal to the signal length 
        """
        if frame_size > np.shape(self.__raw)[1]:
            raise Exception( f"Cannot set frame_size: actual signal length ({np.shape(self.__raw)[1]}) is shorter than frame size ({frame_size}) " )

        self.__frame_size = frame_size
        self.__frame_number = int( np.shape(self.__raw)[1] / self.__frame_size )



def generate_video( imgs: list, rate: float, directory=None, norm: str|None=None, flip=True ):
    """
    Generate a video from a list of images
    
    Parameters
    ----------
    imgs: list
        list of images saved as 2D numpy array
    rate: float
        images number per second (video frequency)
    directory: str, optional
        directory where to save the video
    flip: bool, optional
        flip images vertically
    """

    if directory is None:
        directory = '.'

    log.info( f' .Generate grescale and colored video from energy-type images' )

    if norm is None:        
        # Normalize images in greyscale values
        normalized_imgs = [(img - np.min(img)) / (np.max(img)-np.min(img)) for img in imgs]
    elif norm == 'energy':
        min_pixel_value = min( np.min(img) for img in imgs )
        max_pixel_value = max( np.max(img) for img in imgs )
        dynamic_range = max_pixel_value - min_pixel_value
        normalized_imgs = [(img - min_pixel_value) / (dynamic_range) for img in imgs]
    else:
        raise MuException( f"Unknown normalization method: '{norm}'.")
    
    # Buil grey scale image
    normalized_imgs = [(img * 255).astype(np.uint8) for img in normalized_imgs]

    # Flip images along the vertical axis
    if flip:
        normalized_imgs = [np.flipud(img) for img in normalized_imgs]

    # Convert normalized grayscale images into color images using one of the available colormaps
    cmap = plt.get_cmap('hot')

    # Convert grayscale images to color images
    color_imgs = [cmap(img) for img in normalized_imgs]

    # Save in directory
    mimsave(f'{directory}/video.mp4', normalized_imgs, fps=rate)
    mimsave(f'{directory}/video-colored.mp4', color_imgs, fps=rate)

    log.info( f' .Video saved' )


def generate_movie( imgs: np.ndarray, rate: float, sound: np.ndarray, sampling_frequency: float, norm=str|None, directory=None, colored=True ):
    """
    Generate a film by adding audio to image sequence.
    Images files are build in a ./tmp local directory and removed if `cleanup` is set du True

    Parameters
    ----------
    imgs: np.ndarray
        list of images saved as 2D numpy array
    rate: float 
        images number per second (video frequency)
    sound: np.ndarray
        sound as a numpy array of float
    sampling_frequency: float
        sound sampling frequency
    norm: str, optional
        images normalization method. Can be either None or "energy"
    extend: floats (left, right, bottom, top), optional
        The bounding box in data coordinates that the image will fill.
        The image is stretched individually along x and y to fill the box.
    directory: str, optional
        directory where to save the movie
    cleanup: bool, optional
        clean temporary directory
    """
    
    if directory is None:
        directory = '.'

    # Can work with int type for sampling_frequency
    sampling_frequency = int( sampling_frequency )

    # Create the video from images
    log.info( f' .Generate video from images...' )
    generate_video( imgs=imgs, rate=rate, directory=directory, norm=norm, flip=True )

    # Save sound
    log.info( f' .Generate sound wav file...' )
    wavfile.write ( f"{directory}/audio.wav", sampling_frequency, np.array( sound >> 8 ).astype(np.int16)*4 )

    # Create moovie by merging video and sound
    log.info( f' .Merge audio with video and make mp4 movie file...' )
    if os.path.exists(f'{directory}/movie.mp4'):
        os.remove( f'{directory}/movie.mp4' )
    
    if colored:
        cmd = f"cd {directory} && ffmpeg -v error -i video-colored.mp4 -i audio.wav -map 0:v -map 1:a -c:v copy -shortest movie.mp4"
    else:
        cmd = f"cd {directory} && ffmpeg -v error -i video.mp4 -i audio.wav -map 0:v -map 1:a -c:v copy -shortest movie.mp4"
    error = os.system( cmd )
    if error:
        raise MuException( "failed to write mp4 movie file..." )
    


def generate_moovie_old( imgs: np.ndarray, rate: float, sound: np.ndarray, sampling_frequency: float, norm=str|None, extent=None, directory=None, cleanup=True ):
    """
    This function is deprecated. Use `generate_movie` instead.
    Generate a film by adding audio to image sequence.
    Images files are build in a ./tmp local directory and removed if `cleanup` is set du True

    Parameters
    ----------
    imgs: np.ndarray
        list of images saved as 2D numpy array
    rate: float 
        images number per second (video frequency)
    sound: np.ndarray
        sound as a numpy array of float
    sampling_frequency: float
        sound sampling frequency
    norm: str, optional
        images normalization method. Can be either None or "energy"
    extend: floats (left, right, bottom, top), optional
        The bounding box in data coordinates that the image will fill.
        The image is stretched individually along x and y to fill the box.
    directory: str, optional
        directory where to save the movie
    cleanup: bool, optional
        clean temporary directory
    """
    
    if directory is None:
        directory = '.'

    # Can work with int type for sampling_frequency
    sampling_frequency = int( sampling_frequency )

    # Create tmp directory
    log.info( f' .Create {directory}/tmp directory...' )
    os.system( f'mkdir -p {directory}/tmp && rm -Rf {directory}/tmp/*' )

    # Create video from images 
    if norm == None:
        log.info( f' .Generate images as png files without normalization...' )
        for i, img in enumerate( imgs ):
            plt.imshow( img, origin='lower', extent=extent)
            plt.savefig( f"{directory}/tmp/file{i:02d}.png" )
    elif norm == 'energy':
        log.info( f' .Generate images as png files with sequence energy normalization...' )
        log.info( f' .Found min/max images values in sequence: [{np.amin(imgs)}, {np.amax(imgs) }]' )
        for i, img in enumerate( imgs ):
            plt.imshow( img, vmin=np.amin(imgs), vmax=np.amax(imgs), origin='lower', extent=extent )
            plt.savefig( f"{directory}/tmp/file{i:02d}.png" )
    else:
        raise MuException( f"Unknown normalization method: '{norm}'.")

    # write video
    log.info( f' .Generate video from png files...' )
    cmd = f"cd {directory}/tmp && ffmpeg -v error -r {rate} -i file%02d.png -vcodec mpeg4 -y video.mp4"
    error = os.system( cmd )
    if error:
        raise MuException( "failed to write mp4 video file from png images..." )

    # Save sound
    log.info( f' .Generate sound wav file...' )
    log.info( f' .Sound sampling frequency: {sampling_frequency} Hz' )
    log.info( f' .Sound length: {len(sound)} samples' )
    log.info( f' .Samples format: {sound.dtype}' )
                                            
    wavfile.write ( f"{directory}/tmp/audio.wav", sampling_frequency, sound )

    # merge video and sound
    log.info( f' .Merge audio with video and make mp4 movie file...' )
    cmd = f"cd {directory}/tmp && ffmpeg -v error -i video.mp4 -i audio.wav -map 0:v -map 1:a -c:v copy -shortest movie.mp4"
    error = os.system( cmd )
    if error:
        raise MuException( "failed to write mp4 movie file..." )
    
    log.info( f' .Movie saved' )

    # Remove png files
    if cleanup:
        log.info( f' .Remove temporary png files...' )
        cmd = f"cd {directory}/tmp && rm *.png"
        if os.system( cmd ):
            raise MuException( "failed to cleanup temporary directory" )


# megamicros.tools.geometry.py python software for data processing
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""
This module provides a set of tools for defining antenna geometries.

Features:
    - Create circular antenna arrays
    - Create rectangular antenna arrays
    - Generate 3D coordinates for antenna elements
    
Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""

import numpy as np


def circle( points_number: int, radius: float, height: float, angle_offset: float, clockwise: bool ) -> list:
    """ Create a circle of MEMs positions

    Parameters
    ----------
    points_number : int
        The number of MEMs positions
    radius : float
        The radius of the circle
    height : float
        The height of the antenna
    angle_offset : float
        The angle offset in radians
    clockwise : bool
        The MEMs positions direction

    Returns
    -------
    mems_positions : list
        The MEMs positions list
    """

    mems_positions = [ [0, 0, 0] for i in range( points_number ) ]

    direction = -1. if clockwise else 1.
    for i in range( points_number ):
        angle = direction * ( 2 * np.pi * i / points_number + angle_offset ) + np.pi / 2
        x = radius * np.cos( angle )
        y = radius * np.sin( angle )
        z = height
        mems_positions[i] = [x, y, z]

    return mems_positions

def horizontalPlan( width: float, depth: float, height: float, n_width: int, n_depth: int ) -> list:
    """ Create a horizontal plan of locations

    Parameters
    ----------
    width : float
        The width of the plan
    depth : float
        The depth of the plan
    height : float
        The height of the plan
    n_width : int
        The number of MEMs positions in width
    n_depth : int
        The number of MEMs positions in depth

    Returns
    -------
    locations : list
        The locations list
    """

    locations = [ [0, 0, 0] for i in range( n_width * n_depth ) ]

    step_width = width / n_width
    step_depth = depth / n_depth

    for i in range( n_depth ):
        for j in range( n_width ):
            x = j * step_width - width/2
            y = i * step_depth - depth/2
            z = height
            locations[i*n_width + j] = [x, y, z]

    return locations



# megamicros.acoustics.bmf.py
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Define beamformer classes for beamforming usages

Features:
    - Abstract base class for beamformers
    - Frequency Domain Delay and Sum Beamformer Algorithm (BeamformerFDAS)
    - Several methods to compute the BMF:
        * `type='full'`: beamforming is computed over all the frequencies
        * `type='max'`: beamforming is only performed on the frequency givng the maximum energy (energy is computed on the first MEMs)
        * `type='mean'`: the power spectrum along MEMs 0 is viewed as a probability density which is approximated by a gaussian distribution. Beamforming is computed on the mean frequency of the gaussian distribution.
        * `type='gauss'`: the power spectrum along MEMs 0 is viewed as a probability density which is approximated by a gaussian distribution. Beamforming is computed on a range of frequencies around the mean frequency of the gaussian distribution with standard deviation as the width of the range.
        * `type='omp'`: find the location that best matches the phase distribution at the frequency for which energy is max using OMP algorithm
        
Examples:
    Basic usage::

        Comming soon...

    Advanced usage::

        See the Notebooks for advanced usage examples.

Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""

import numpy as np

from ..exception import MuException
from ..log import log
from ..acoustics.omp import omp, Result


SOUND_SPEED = 340.29

class MuBmfException( MuException ):
    """ Exception for beamformers """
    pass


class Beamformer:
    """ Abstract base class for beamformers """

    __mems_position: np.ndarray                         # 3D MEMs absolute positions in meters
    __locations: np.ndarray                             # 3D absolute positions of locations in the space where the beamforming is computed
    __sampling_frequency: float                         # Input sampling frequency
    __frame_length: int                                 # Input frame length in samples

    @property
    def mems_position( self ) -> np.ndarray:
        """ Get 3D absolute MEMS positions """

        return self.__mems_position

    @property
    def locations( self ) -> np.ndarray:
        """ Get 3D absolute location's positions """
        
        return self.__locations

    @property
    def sampling_frequency( self ) -> float:
        """ Sampling frequency """
        
        return self.__sampling_frequency
    
    @property
    def frame_length( self ) -> int:
        """ Sampling frequency """
        
        return self.__frame_length
    
    @property
    def mems_number( self ) -> int:
        """ MEMs number according the mems_position array """
        
        return np.shape( self.__mems_position )[0]
    
    @property
    def locations_number( self ) -> int:
        """ Locations number according the locations array """
        
        return np.shape( self.__locations )[0]
    

    def setMemsPosition( self, mems_position: np.ndarray|list|tuple ) -> None:
        """ Set the 3D absolute MEMs positions """

        if type( mems_position is list or type( mems_position is tuple ) ):
            mems_position = np.array( mems_position )

        if len( mems_position.shape ) != 2:
            raise MuException( f"Cannot set MEMs positions: MEMs positions must be a 2D array (MEMs_number x 3) but found {mems_position.shape}" )
        
        if mems_position.shape[1] != 3:
            raise MuException( f"Cannot set MEMs positions: MEMs positions must be a 3D array (MEMs_number x 3) but found {mems_position.shape}" )
        
        log.info( f" .Set beamformer on a {np.shape( mems_position )[0]} MEMs antenna" )
        self.__mems_position = mems_position


    def setLocations( self, locations: np.ndarray|list|tuple ) -> None:
        """ Set the space locations where to compute the BMF """

        if type( locations is list or type( locations is tuple ) ):
            locations = np.array( locations )

        if len( locations.shape ) != 2:
            raise MuException( f"Cannot set space locations: space locations must be a 2D array (locations_number x 3) but found {locations.shape}" )
        
        if locations.shape[1] != 3:
            raise MuException( f"Cannot set space locations: space locations must be a 3D array (MEMs_number x 3) but found {locations.shape}" )
        
        log.info( f" .Set {np.shape( locations )[0]} beamforming locations" )
        self.__locations = locations


    def setSamplingFrequency( self, sampling_frequency: float ) -> None :
        """ Set the sampling frequency of input signals """

        log.info( f" .Set beamformer sampling frequency to {sampling_frequency} Hz" )
        self.__sampling_frequency = sampling_frequency


    def setFrameLength( self, frame_length: float ) -> None :
        """ Set the input data buffer length in samples number """

        log.info( f" .Set beamformer frame length to {frame_length} Samples" )
        self.__frame_length = frame_length


    def __init__( self, mems_position: np.ndarray, locations: np.ndarray, sampling_frequency: float, frame_length: int ):
        """ Create a new beamformer instance 
        
        Parameters
        ----------
        mems_positions: np.ndarray
            The MEMs positions as a 2D array (MEMs_number x 3)
        locations: np.ndarray
            The space locations where to compute the BMF
        sampling_frequency: float
            The sampling frequency of input signals
        frame_length: int
            The frame length in samples number
        """

        self.setMemsPosition( mems_position )
        self.setSamplingFrequency( sampling_frequency )
        self.setLocations( locations )
        self.setFrameLength( frame_length )


class BeamformerFDAS( Beamformer ):
    """ Frequency Domain Delay and Sum Beamformer Algorithm """

    __f_axis: np.ndarray                        # Frequency axis in Hertz
    __fft_low_cut_off: float                    # FFT low cut off frequency
    __fft_high_cut_off: float                   # FFT high cut off frequency
    __fft_window_size: int                      # FFT window size in samples number

    __mems_number: int                          # MEMs number according the mems_position array
    __band_width: tuple = [0, 1]                # Normalized frequencies bandwidth
    __fft_low_cut_off_index: int                # Bandwidth start frequency range idex
    __fft_high_cut_off_index: int               # Bandwidth end frequency range index
    __band_width_length: int                    # Bandwidth length in samples number

    __D: np.ndarray                             # Inter mems/locations distance matrix
    __H: np.ndarray                             # beamforming matrix
    __BFE: np.ndarray                           # Beamforming energy
    __BFSpec: np.ndarray                        # Complex spectrum computed on all beams (locations_number x frequencies_number)
    __FFT: np.ndarray                           # Complex spectrum computed on all channels (channels_number x frequencies_number)
    
    @property
    def BFE( self ) -> np.ndarray:
        """ Beamforming energy array """

        return self.__BFE
    
    @property
    def D( self ) -> np.ndarray:
        """ Distances matrix """

        return self.__D

    @property
    def H( self ) -> np.ndarray:
        """ Beamforming matrix """

        return self.__H
    
    @property
    def SV( self, freq, pos ) -> np.ndarray:
        """ Steering vector for a given position and frequency """

        return self.__H[freq, pos, :]

    @property
    def FFT( self ) -> np.ndarray:
        """ FFT matrix """

        return self.__FFT

    def getBeamformingEnergy( self ) -> np.ndarray:
        """ Get the beamforming energy array """

        return self.__BFE

    def getFrequenciesAxis( self ) -> np.ndarray:
        """ Get the frequencies axis """

        return self.__f_axis

    def setFFtWindowSize( self, fft_window_size):
        """ Set the FFT window size in samples number 
        
        Note that it is not required that the FFT window size is the same as frame length. But only same size is allowed for now
        """
        if fft_window_size != self.frame_length:
            raise MuBmfException( f"Cannot set FFT window size to {fft_window_size} samples. Should be same as frame length ({self.frame_length} samples)" )

        log.info( f" .Set beamformer FFT window size to {fft_window_size} samples" )
        self.__fft_window_size = fft_window_size


    def __init__( self, mems_position: np.ndarray, locations: np.ndarray, sampling_frequency: float, frame_length: int, bandwidth: tuple|None=None ):
        """ Create a new Frequency Domain Adaptive Beamformer by delay and sum method instance 
        Parameters
        ----------
        mems_positions: np.ndarray
            The MEMs positions as an array of 3D array positions (MEMs number x 3)
        locations: np.ndarray
            The space locations where to compute the BMF as an array of 3D array positions (locations number x 3)
        sampling_frequency: float
            The sampling frequency of input signals
        frame_length: int
            The frame length in samples number
        bandwidth: tuple, optional
            The normalized frequencies bandwidth. Default is [0, 1] which means all frequencies from 0 to Fe/2
        """
        
        super().__init__( mems_position, locations, sampling_frequency, frame_length )

        locations_number = np.shape( self.locations )[0]
        self.setFFtWindowSize( self.frame_length )

        # time axis in seconds
        t = np.arange( self.__fft_window_size )/self.sampling_frequency

        # frequency axis in Hz
        self.__f_axis = np.fft.rfftfreq( self.__fft_window_size, 1/self.sampling_frequency )

        # frequencies number
        freq_number = self.__f_axis.size

        # frequency step
        frequency_step = self.sampling_frequency / freq_number / 2

        # bandwidth range
        if bandwidth is not None:
            self.__band_width = bandwidth
        else:
            self.__band_width = [0, 1]
        
        self.__fft_low_cut_off_index = int( self.sampling_frequency * self.__band_width[0] / frequency_step / 2 )
        self.__fft_high_cut_off_index = int( self.sampling_frequency * self.__band_width[1] / frequency_step / 2 ) - 1
        self.__band_width_length = self.__fft_high_cut_off_index - self.__fft_low_cut_off_index + 1
        self.__fft_low_cut_off = self.__fft_low_cut_off_index * frequency_step
        self.__fft_high_cut_off = (self.__fft_high_cut_off_index+1) * frequency_step

        # print info
        log.info( f" .BeamformerFDAS Initilization:" )
        log.info( f"  > Found antenna with {self.mems_number} MEMs microphones" )
        log.info( f"  > FFT window size is {self.__fft_window_size} samples" )
        log.info( f"  > Time range: [0, {t[-1]}] s" )
        log.info( f"  > Sampling frequency: {self.sampling_frequency} Hz" )
        log.info( f"  > Frequency range: [0, { self.__f_axis[-1]}] Hz ({freq_number} beams)" )
        log.info( f"  > frequency step: {frequency_step:.2f} Hz" )
        log.info( f"  > frequency bandwidth: [{self.__fft_low_cut_off:.2f}, {self.__fft_high_cut_off:.2f}] Hz" )
        log.info( f"  > frequency bandwidth indexes: [{self.__fft_low_cut_off_index}, {self.__fft_high_cut_off_index}] ( {self.__band_width_length} spectral ray)" )

        # Init distance matrix
        log.info( f" .Build distances matrix D ({locations_number} x {self.mems_number})" ) 
        self.__D = np.ndarray( (locations_number, self.mems_number), dtype=float )
        for s in range( locations_number ):
            for m in range( self.mems_number ):
                self.__D[s, m] = np.linalg.norm( np.array( self.mems_position[m] ) - self.locations[s] )

        # Allocate and build the H complex transfer function matrix (preformed channels)
        log.info( f" .Build preformed channels matrix H ({freq_number} x {locations_number} x {self.mems_number})" ) 
        self.__H = np.outer(  self.__f_axis, self.__D ).reshape( freq_number, locations_number, self.mems_number )/SOUND_SPEED
        self.__H = np.exp( 1j*2*np.pi*self.__H )

        self.__FFT = None


    def compute( self, signal: np.ndarray, type='full' ) -> np.ndarray:
        """ Process beamforming on input signals

        If the signal length is smaller than the `__fft_window_size` parameter, the input is zero padded.
        If it is larger, the input signal is cropped.
        
        There are several methods to compute the BMF:
        * `type='full'`: beamforming is computed over all the frequencies
        * `type='max'`: beamforming is only performed on the frequency givng the maximum energy (energy is computed on the first MEMs)
        * `type='mean'`: the power spectrum along MEMs 0 is viewed as a probability density which is approximated by a gaussian distribution. Beamforming is computed on the mean frequency of the gaussian distribution.
        * `type='gauss'`: the power spectrum along MEMs 0 is viewed as a probability density which is approximated by a gaussian distribution. Beamforming is computed on a range of frequencies around the mean frequency of the gaussian distribution with standard deviation as the width of the range.
        * `type='omp'`: find the location that best matches the phase distribution at the frequency for which energy is max using OMP algorithm
        
        Parameters
        ----------
        signal: np.ndarray
            the MEMs signal line wise (samples_number X mems_number)

        Return
        ------
        BFE: np.ndarray
            The beamformed energy channels (location_number x 1)
        """
        signal_length, mems_number = signal.shape
        # Control input size
        if signal_length > self.__fft_window_size:
            log.warning( f" .bmf > Input signal is longer than FFT width: it will be truncated" )
            self._in = signal[:self.__fft_window_size,:]
        elif signal_length < self.__fft_window_size:
            log.warning( f" .bmf > Input signal is shorter than FFT width: it will be zero-padded" )
            self._in = np.pad( signal, ( ( 0,self.__fft_window_size-signal_length), (0,0) ) )

        # Process beamforming on input signal involving all frequencies
        if type == 'full':
            self.__FFT = np.fft.rfft( signal, n=self.__fft_window_size, axis=0 )
            # SpecH = self.__FFT[:, None, :] * self.__H
            # BFSpec = np.sum( SpecH, -1 ) / self.mems_number
            # self.__BFE = np.sum( ( np.abs( BFSpec )**2 )[self.__fft_low_cut_off_index:self.__fft_high_cut_off_index+1,:], 0 ) / self.__band_width_length

            SpecH = self.__FFT[self.__fft_low_cut_off_index:self.__fft_high_cut_off_index, None, :] * self.__H[self.__fft_low_cut_off_index:self.__fft_high_cut_off_index,:,:]
            BFSpec = np.sum( SpecH, -1 ) / self.mems_number
            # self.__BFE = np.sum( np.abs( BFSpec )**2, 0 ) / self.__band_width_length
            self.__BFE = np.sum( np.abs( BFSpec ), 0 ) / self.__band_width_length
            self.__BFE = self.__BFE**2
            # Compute power in db:
            self.__BFE = 10 * np.log(self.__BFE)

            return self.__BFE, None, None

        # compute beamforming on frequency with max energy
        elif type == "max":
            # compute spectrum on all channels            
            self.__FFT = np.fft.rfft( signal, n=self.__fft_window_size, axis=0 )

            # find the max energy on first MEMs microphone
            #freq_max_index = np.argmax( np.abs( Spec[:,0] )**2 )
            freq_max_index = np.argmax( np.mean( np.abs( self.__FFT ), axis=1 ) )

            # compute bmf on the selected frequency
            SpecH = self.__FFT[freq_max_index, None, :] * self.__H[freq_max_index,:,:]
            BFSpec = np.sum( SpecH, -1 ) / self.mems_number

            self.__BFE = np.abs( BFSpec )**2 / BFSpec.size
            selected_frequency = self.__f_axis[freq_max_index]
            band_width = None
            return self.__BFE, selected_frequency, band_width

        # compute beamforming on the mean energy frequency
        elif type == "mean":
            # compute spectrum on all channels            
            self.__FFT = np.fft.rfft( signal, n=self.__fft_window_size, axis=0 )

            # find the mean energy frequency on channel 0
            Spec0 = np.abs( self.__FFT[:,0] )**2
            Spec0_density = Spec0 / np.sum( Spec0 )
            freq_mean = np.sum(  self.__f_axis * Spec0_density )

            # find the nearest frequency in the FFT axis
            freq_mean_index = np.searchsorted( self.__f_axis, freq_mean )

            # Get the nearest frequency index
            if freq_mean_index > 0:
                if np.abs(self.__f_axis[freq_mean_index-1] - freq_mean) < np.abs(self.__f_axis[freq_mean_index] - freq_mean):
                    freq_mean_index = freq_mean_index - 1
            
            # compute bmf on the selected frequency
            SpecH = self.__FFT[freq_mean_index, None, :] * self.__H[freq_mean_index,:,:]
            BFSpec = np.sum( SpecH, -1 ) / self.mems_number
            self.__BFE = np.abs( BFSpec )**2 / BFSpec.size
            selected_frequency = self.__f_axis[freq_mean_index]
            band_width = None


        # compute beamforming on frequencies around the mean energy frequency bounded by standard deviation
        elif type == "gauss":
            # compute spectrum on all channels            
            self.__FFT = np.fft.rfft( signal, n=self.__fft_window_size, axis=0 )

            # find the mean energy frequency on channel 0
            Spec0 = np.abs( self.__FFT[:,0] )**2
            Spec0_density = Spec0 / np.sum( Spec0 )
            freq_mean = np.sum(  self.__f_axis * Spec0_density )
            std_dev = np.sqrt( np.sum( (self.__f_axis - freq_mean)**2 * Spec0_density ) ) / 2

            # find the nearest frequency in the FFT axis
            freq_mean_index = np.searchsorted( self.__f_axis, freq_mean )

            # Get the nearest frequency index
            if freq_mean_index > 0:
                if np.abs(self.__f_axis[freq_mean_index-1] - freq_mean) < np.abs(self.__f_axis[freq_mean_index] - freq_mean):
                    freq_mean_index = freq_mean_index - 1

            # Find bandwith correponding to standard deviation
            Df = self.__f_axis[1] - self.__f_axis[0]
            std_dev_nindex = int( std_dev / Df )

            freq_min_index = max( freq_mean_index - std_dev_nindex, 0 )
            freq_max_index = min( freq_mean_index + std_dev_nindex, self.__f_axis.size )
            
            # compute bmf on the selected frequency
            SpecH = self.__FFT[freq_min_index:freq_max_index, None, :] * self.__H[freq_min_index:freq_max_index,:,:]
            BFSpec = np.sum( SpecH, -1 ) / self.mems_number
            self.__BFE = np.sum( np.abs( BFSpec )**2, 0 ) / BFSpec.size
            selected_frequency = self.__f_axis[freq_mean_index]
            band_width = std_dev


        # Find the location that best matches the phase distribution at the frequency for which energy is max
        elif type == "omp":
            # compute spectrum on all channels            
            self.__FFT = np.fft.rfft( signal, n=self.__fft_window_size, axis=0 )

            # init result vector
            omp_fo = np.zeros( self.locations_number )

            # frequency index carrying max energy
            freq_max_index = np.argmax( np.mean( np.abs( self.__FFT ), axis=1 ) )

            # Set dictionnary as the set of all phases complex multiplicators at the i_f0 frequency 
            Dico = self.__H[freq_max_index,:,:].T

            # Compute the first decomposition coef given by matching pursuit 
            result: Result = omp( Dico, self.__FFT[freq_max_index,:], maxit=1, verbose=False )

            # Get coef in the location vector (that could be reshaped as (nx, ny))  
            omp_fo = result.coef + 1e-12
            self.__BFE = omp_fo

            selected_frequency = self.__f_axis[freq_max_index]
            band_width = None


        return self.__BFE, selected_frequency, band_width


    def plotBfe( self, room_width: float, room_depth: float, x_npos: float, y_npos: float, position=None, prediction=None ):
        """ Plot the BFE in a room and highlight the max value and the source position"""

        bfe_room = np.reshape( self.__BFE, ( x_npos, y_npos ) )

        fig, ax = plt.subplots()
        img = ax.imshow( bfe_room, origin='lower' )
        if prediction is not None:
            ax.set_title( f"BFE - max found at ({prediction[0]:.2f},{prediction[1]:.2f}) meters" )

        xticks = np.linspace( 0, x_npos-1, 10 )
        xticks_number = len( xticks )
        yticks = np.linspace( 0, y_npos-1, 10 )
        yticks_number = len( yticks )
        ax.set_xticks( xticks, labels=np.array( [i for i in range( xticks_number )] )*room_width//(xticks_number-1) )
        ax.set_yticks( yticks, labels=np.array( [i for i in range( yticks_number )] )*room_depth//(yticks_number-1) )

        if position is not None:
            ax.scatter( position[0]*x_npos/room_width, position[1]*y_npos/room_depth, color='green' )

        if prediction is not None:
            ax.scatter( prediction[0]*x_npos/room_width, prediction[1]*y_npos/room_depth, color='orange' )

        fig.colorbar( img, ax=ax, label='Interactive colorbar' )

        return ax


# megamicros.usb.py
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Define USB device handling functions to communicate with Megamicros hardware.

Features:
    - Handle USB device connection, claiming, and release
    - Perform synchronous and asynchronous bulk read transfers
    - Support for control write commands to the USB device

Basic usage::

    try:
        if Usb.checkDeviceByVendorProduct( vendor_id=0xFE27, product_id=0xAC03 ):
            print( "Device found!" )
            usb_device = Usb()
            usb_device.open( vendor_id=0xFE27, product_id=0xAC03, bus_address=0x00, endpoint_in=0x81 )
            usb_device.claim()
            usb_device.release()
            usb_device.close()
        else:
            print( "Device not found!" )
    except Exception as e:
        print( f"An error occurred: {e}" )
    
Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""

import queue
from ctypes import addressof, byref, sizeof, create_string_buffer, CFUNCTYPE
import usb1
import threading

from .log import log
from .exception import MuException

# libusb constants (provided by the C libusb but not accessible through python/usb1)
LIBUSB_RECIPIENT_DEVICE      = 0x00             
LIBUSB_REQUEST_TYPE_VENDOR   = 0x40
LIBUSB_ENDPOINT_OUT          = 0x00

# Default constants to perform input bulk transfers from the usb device
USB_DEFAULT_INTERFACE        = 0x00
USB_DEFAULT_ENDPOINT_IN      = 0x81
USB_DEFAULT_TRANSFER_TIMEOUT  = 1000                # timeout in ms
USB_DEFAULT_WRITE_TIMEOUT     = 1000                # timeout in ms

USB_DEFAULT_BUFFERS_NUMBER   = 8                    # Default usb transfer buffers number
USB_DEFAULT_QUEUE_SIZE       = 0                    # Queue limit after which the last signal is lost (0 means infinite signal queueing)
USB_DEFAULT_QUEUE_TIMEOUT    = 1000                 # Queue get timeout in ms


                                                                            
class UsbException( MuException ):
    """ Exception base class for USB devices in Megamicros """

    def __init__( self, message: str="" ):
        super().__init__( message )

class Usb:
    """ Class representing a USB device """

    class Queue( queue.Queue ):
        """ Thread safe queue adapted to the Megamicros USB class 
        
        Original comportment is that of blocks insertion once maxsize is reached, until queue items are consumed.
        This implementation pop up the last element of the queue if maxsize is reached.
        """

        @property
        def transfert_lost( self ) -> bool:
            """ Get the transfer lost counter """
            return self.__transfert_lost
    
        def __init__( self, maxsize: int = 0 ):
            self.__transfert_lost: int = 0
            super().__init__( maxsize )

        def clear( self ) -> None:
            """ Clear the queue """
            # Manually empty the queue since Queue doesn't have a clear() method
            while not self.empty():
                try:
                    self.get_nowait()
                except queue.Empty:
                    break
            self.__transfert_lost = 0

        def put( self, data ):
            # pop up the last element of the queue if maxsize is reached.
            if self.maxsize > 0 and self.qsize() >= self.maxsize:
                self.get()
                self.__transfert_lost += 1

            # Let parent class do the work 
            super().put( data )


    def __init__( self, vendor_id:int|None = None, product_id:int|None = None, bus_address:int|None = None, endpoint_in:int|None = None, endpoint_out:int|None = None ):
        self.__vendor_id = vendor_id
        self.__product_id = product_id
        self.__bus_address = bus_address
        self.__endpoint_in = endpoint_in
        self.__endpoint_out = endpoint_out
        self.__context: usb1.USBContext|None = None
        self.__usb_handle: usb1.USBDeviceHandle|None = None
        self.__is_open = False
        self.__is_claimed = False
        self.__buffer_size: int = 0
        self.__buffers_number: int = USB_DEFAULT_BUFFERS_NUMBER

        # Transfer parameters
        self.__transfer_timeout: int = USB_DEFAULT_TRANSFER_TIMEOUT
        self.__transfer_buffers: list[usb1.USBTransfer] = []
        self.__bulk_transfer_on: bool = False
        self.__on_stop_callback: callable|None = None
        self.__async_transfer_thread: threading.Thread|None = None
        self.__async_transfer_thread_exception: Exception|None = None

        # Timer management
        self.__timer_thread: threading.Timer|None = None

        # Queue management
        self.__queue_size: int = USB_DEFAULT_QUEUE_SIZE
        self.__queue: Usb.Queue = Usb.Queue( maxsize=USB_DEFAULT_QUEUE_SIZE )

        if vendor_id is not None and product_id is not None and bus_address is not None :
            self.open( vendor_id, product_id, bus_address, endpoint_in if endpoint_in is not None else USB_DEFAULT_ENDPOINT_IN, endpoint_out )

    def __del__( self ):
        self.close()
        log.info( ' .USB object destroyed' )

    @property
    def buffer_size( self ) -> int:
        return self.__buffer_size

    @property
    def buffers_number( self ) -> int:
        return self.__buffers_number

    @property
    def transfer_timeout( self ) -> int:
        return self.__transfer_timeout

    @property
    def bus_address( self ) -> int:
        return self.__bus_address
    
    @property
    def queue( self ) -> 'Usb.Queue':
        return self.__queue
    
    @property
    def queue_size( self ) -> int:
        return self.__queue_size
    
    @property
    def queue_content( self ) -> int:
        return self.__queue.qsize()

    @property
    def transfer_lost( self ) -> int:
        return self.__queue.transfert_lost

    def isOpened( self ) -> bool:
        """ Check if the USB device is open """
        return self.__is_open

    @staticmethod
    def checkDeviceByVendorProduct( vendor_id:int, product_id:int ) -> bool:
        """ 
        Check if the connected USB device matches the given vendor_id and product_id

        Parameters
        ----------
        vendor_id: int
            The vendor ID to check
        product_id: int
            The product ID to check

        Returns
        -------
        bool
            True if the connected USB device matches the given vendor_id and product_id, False otherwise
        """

        context = usb1.USBContext()
        usb_handle = context.openByVendorIDAndProductID( 
            vendor_id, 
            product_id,
            skip_on_error=True,
        )

        if usb_handle is None:
            return False
        else:
            usb_handle.close()
            context.close()
            return True



    def setBufferSize( self, size:int ) -> None:
        """ Set the USB transfer buffer size in bytes"""
        self.__buffer_size = size

    def setBuffersNumber( self, number:int ) -> None:
        """ Set the USB transfer buffers number """
        self.__buffers_number = number

    def setQueueSize( self, size:int ) -> None:
        """ Set the USB transfer queue size in bytes (0 means infinite queueing) 
        Queue is cleared when size is changed
        """
        self.__queue_size = size
        self.__queue = Usb.Queue( maxsize=size )

    def setOnStopCallback( self, callback: callable ) -> None:
        """ Set the callback function to be called when the asynchronous bulk transfer stops """
        self.__on_stop_callback = callback

    def open( self, vendor_id:int, product_id:int, bus_address:int, endpoint_in:int, endpoint_out:int|None = None ) -> None:
        """ 
        Open the USB device connection - but don't claim it yet 
        Try to locate the device based on vendor_id, product_id, bus_address
        """

        log.info( f' .Connecting to USB device {vendor_id:04x}:{product_id:04x} ...' )
        self.__context = usb1.USBContext()
        self.__usb_handle = self.__context.openByVendorIDAndProductID( 
            vendor_id, 
            product_id,
            skip_on_error=True,
        )

        if self.__usb_handle is None:
            raise UsbException( 'Failed to connect to USB device: the device may be disconnected or user not allowed to access' )

        self.__vendor_id = vendor_id
        self.__product_id = product_id
        self.__bus_address = bus_address
        self.__endpoint_in = endpoint_in
        self.__endpoint_out = endpoint_out

        # Test claiming the interface:
        self.claim()

        # release the interface
        self.release()

        self.__is_open = True
        log.info( f' .Connected on USB device {self.__vendor_id:04x}:{self.__product_id:04x}' )


    def claim(self) -> None:
        """ Claim the USB device interface """
        if self.__is_claimed == False:
            if self.__usb_handle is not None:
                if self.__usb_handle.claimInterface( self.__bus_address ) == False:
                    raise UsbException( f'Failed to claim interface {self.__bus_address} on USB device {self.__vendor_id:04x}:{self.__product_id:04x}' )
                log.info( f' .Claimed interface {self.__bus_address} on USB device {self.__vendor_id:04x}:{self.__product_id:04x}' )
                self.__is_claimed = True

    def release(self) -> None:
        """ Release the USB device interface """
        if self.__is_claimed:
            if self.__usb_handle is not None:
                if self.__usb_handle.releaseInterface( self.__bus_address ) == False:
                    raise UsbException( f'Failed to release interface {self.__bus_address} on USB device {self.__vendor_id:04x}:{self.__product_id:04x}' )
                log.info( f' .Released interface {self.__bus_address} on USB device {self.__vendor_id:04x}:{self.__product_id:04x}' )
                self.__is_claimed = False

    def close( self ) -> None:
        """ Close the USB device connection """
        if self.__is_open:
            if self.__is_claimed:
                self.release()
            if self.__usb_handle is not None:
                self.__usb_handle.close()
            if self.__context is not None:
                self.__context.close()
            self.__is_open = False
            log.info( f' .Disconnected from USB device {self.__vendor_id:04x}:{self.__product_id:04x}' )

    def ctrlWrite( self, request, data, time_out=USB_DEFAULT_WRITE_TIMEOUT ) -> None:
        """
        Send a write command to the USB interface

        Parameters
        ----------
        request: int
            The request code to send
        data: bytes
            The data to send
        time_out: int, optional
            The USB command timeout in ms (default is 1000ms)
        recipient_device: int, optional
            The USB recipient device (default is 0x00)
        type_vendor: int, optional
            The USB request type (default is 0x40)
        endpoint_out: int, optional
            The USB endpoint out (default is 0x00)
        """

        if not self.__usb_handle:
            raise UsbException( 'Cannot write on usb device: no available handler. Please open device before using it!' )

        keeping_claimed = self.__is_claimed
        if not self.__is_claimed:
            self.claim()

        try:
            ndata = self.__usb_handle.controlWrite(
                LIBUSB_RECIPIENT_DEVICE | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_ENDPOINT_OUT,  # bmRequestType
                request, 	    # command
                0,			    # command parameter value
                0,			    # index
                data,		    # data to send 
                time_out        # timeout in ms 
            )

            if ndata != sizeof( data ):
                log.warning( ' .In Usb.__ctrlWrite(): Command failed with ', ndata, ' data transfered against ', sizeof( data ), ' wanted ' )

            if not keeping_claimed:
                self.release()

        except Exception as e:
            log.error( f"write failed on device: {e}" )
            raise


    def ctrlWriteReset( self, request, time_out=USB_DEFAULT_WRITE_TIMEOUT ) -> None:
        """
        Send a reset write command to the MegaMicro FPGA through the usb interface.
        This command needs to perform a _controlTransfer() call instead of a controlWrite() call.
        This is because we have no data to transfer (0 length) while the buffer should not be empty.
        controlWrite() computes the data length on its own, that is something >0 leading to a LIBUSB_ERROR_PIPE [-9] exception
        """

        if not self.__usb_handle:
            raise UsbException( 'Cannot write on usb device: no available handler. Please open device before using it!' )

        keeping_claimed = self.__is_claimed
        if not self.__is_claimed:
            self.claim()
        
        data = create_string_buffer( 16 )
        try:
            ndata = self.__usb_handle._controlTransfer(
                LIBUSB_RECIPIENT_DEVICE | LIBUSB_REQUEST_TYPE_VENDOR | LIBUSB_ENDPOINT_OUT,
                request, 
                0,
                0, 
                data, 
                0,
                time_out,
            )
        except Exception as e:
            log.error( f"reset write failed on device: {e}" )
            raise

        if ndata != 0:
            log.warning( ' .In Usb.__ctrlWrite(): Command failed with ', ndata, ' data transfered against 0 wanted' )
        
        if not keeping_claimed:
            self.release()

    def syncBulkRead( self, size:int, time_out=USB_DEFAULT_TRANSFER_TIMEOUT ) -> bytes:
        """
        Perform a synchronous bulk read transfer on the USB device

        Parameters
        ----------
        size: int
            The number of bytes to read
        time_out: int, optional
            The USB command timeout in ms (default is 1000ms)

        Returns
        -------
        bytes
            The data read from the USB device
        """

        if not self.__usb_handle:
            raise UsbException( 'Cannot read from usb device: no available handler. Please open device before using it!' )

        keeping_claimed = self.__is_claimed
        if not self.__is_claimed:
            self.claim()

        try:
            data = self.__usb_handle.bulkRead(
                self.__endpoint_in,
                size,
                time_out,
            )
        except usb1.USBError as e:
            log.error( f"Error during bulk read of {size} bytes from USB device {self.__vendor_id:04x}:{self.__product_id:04x}: {e}" )
            raise UsbException( f"Error during bulk read of {size} bytes from USB device {self.__vendor_id:04x}:{self.__product_id:04x}: {e}" ) from e

        if not keeping_claimed:
            self.release()

        return data
    
    def __callback( self, transfer: usb1.USBTransfer ) -> None:
        """
        Callback function for asynchronous bulk transfer on the USB device
        """

        if transfer.getStatus() == usb1.TRANSFER_COMPLETED:
            data = transfer.getBuffer()[:transfer.getActualLength()]

            # Put data in the queue
            self.__queue.put( data )

            # Resubmit the transfer
            if self.__bulk_transfer_on:
                transfer.submit()

    def asyncBulkTransfer( self, duration: int ) -> None:
        """
        Perform an asynchronous bulk transfer on the USB device.
        This call is non blocking. The device is claimed during all the transfer 
        """

        if not self.__usb_handle:
            raise UsbException( 'Failed to start asynchronous bulk transfer: no available handler. Please open device before using it!' )

        if self.__buffers_number == 0:
            raise UsbException( 'Failed to start asynchronous bulk transfer: buffers number is not set.' )
        
        if self.__buffer_size == 0:
            raise UsbException( 'Failed to start asynchronous bulk transfer: no allocated buffers.' )

        if self.__async_transfer_thread is not None:
            raise UsbException( 'Failed to start asynchronous bulk transfer: another transfer is already running.' )
        
        # Claims the interface
        if not self.__is_claimed:
            self.claim()

        log.info( ' .Init asynchronous bulk transfer...')

        # Clear previous transfer buffers if any
        self.__transfer_buffers.clear()

        # Clear the queue
        self.__queue.clear()

        # Allocate the list of transfer buffers
        for id in range( self.buffers_number ):
            transfer = self.__usb_handle.getTransfer()
            transfer.setBulk(
                usb1.ENDPOINT_IN | self.__endpoint_in,
                self.buffer_size,
                callback=self.__callback,
                user_data = id,
                timeout=self.transfer_timeout
            )
            self.__transfer_buffers.append( transfer )

        # Start the timer if a limited execution time is requested
        # In this case, the transfer ending is scheduled after duration seconds
        if duration > 0 :
            self.__timer_thread = threading.Timer( duration, self.__timer_end_of_transfer_thread )
            self.__timer_thread_flag = True
            self.__timer_thread.start()

        # Start run thread
        self.__async_transfer_thread_exception = None
        self.__async_transfer_thread = threading.Thread( target= self.__asyncBulkTransfer_thread )
        self.__async_transfer_thread.start()

    def __asyncBulkTransfer_thread( self, time_out=USB_DEFAULT_TRANSFER_TIMEOUT ) -> None:
        """
        Thread function to perform an asynchronous bulk transfer on the USB device
        """
        log.info( " .Transfer thread execution started" )

        # Submit all transfers
        for transfer in self.__transfer_buffers:
            transfer.submit()

        # run de main transfer loop
        self.__bulk_transfer_on = True

        # Waits for pending transfers.
        # Once a transfer is finished, handleEvents() triggers the callback function
        try:
            while self.__bulk_transfer_on:
                while any( x.isSubmitted() for x in self.__transfer_buffers ) and self.__bulk_transfer_on:
                    self.__context.handleEvents()

            if not self.__bulk_transfer_on:
                log.info( f" .Quitting transfer loop due to end of process request..." )
            else:
                log.info( f" .Quitting transfer loop due to all transfers completed, timeout reached or error..." )
                self.__asyncBulkTransferStop_thread()
    
        except usb1.USBError as e:
            log.error( f" .Error during asynchronous bulk transfer on USB device {self.__vendor_id:04x}:{self.__product_id:04x}: {e}" )
            log.error( f" .Error resulting in thread termination ({type(e).__name__}): {e}" )
            self.__async_transfer_thread_exception = e
            self.__asyncBulkTransferStop_thread()

    def __asyncBulkTransferStop_thread( self ) -> None:
        """
        Thread function to stop an asynchronous bulk transfer on the USB device
        """
        # Run user callback if any
        if self.__on_stop_callback is not None:
            self.__on_stop_callback()

        self.__bulk_transfer_on = False


    def __timer_end_of_transfer_thread( self ) -> None:
        """ Timer callback for run stopping """

        log.info( f" .Thread timer ended: stop the bulk transfer..." )
        self.__asyncBulkTransferStop_thread()
        self.__thread_timer_flag = False
    
    def asyncBulkTransferWait( self, time_out=USB_DEFAULT_TRANSFER_TIMEOUT ) -> bytes:
        """
        Wait for the end of an asynchronous bulk transfer on the USB device.
        This is a blocking call.
        """

        if self.__async_transfer_thread is not None:
            self.__async_transfer_thread.join()
            self.__async_transfer_thread = None

        if self.__async_transfer_thread_exception is not None:
            thread_exception = MuException( f"Thread exception ({type(self.__async_transfer_thread_exception).__name__}): {self.__async_transfer_thread_exception}" )
            self.__async_transfer_thread_exception = None
            raise thread_exception

    
    def asyncBulkTransferStop( self ) -> None:
        """
        Stop an asynchronous bulk transfer on the USB device
        Not implemented yet
        """
        raise NotImplementedError( 'Usb.asyncBulkTransferStop() not implemented yet' )
    



# megamicros.log.py
#
# ® Copyright 2024-2025 Bimea
# Author: bruno.gas@bimea.io
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Megamicros logging utilities.

This module provides logging functionality for the Megamicros library with colored output
and file logging capabilities.

Features:
    - Stream handler for console output with colors (Unix-like systems)
    - File handler for persistent logging to './megamicros.log'
    - Configurable log levels: debug, info, warning, error, critical
    - Default level set to NOTSET (no output until explicitly configured)

Examples:
    Basic usage::

        from megamicros import log
        
        # Set logging level
        log.setLevel('info')
        
        # Log messages
        log.debug('Debug message')
        log.info('System ready')
        log.warning('This is a warning')
        log.error('An error occurred')
        log.critical('Critical system failure')

    Advanced usage::

        from megamicros import log
        import logging
        
        # Enable debug mode for development
        log.setLevel('debug')
        
        # Check current level
        current_level = log.level
        print(f"Current log level: {current_level}")
        
        # Use with exception handling
        try:
            # Your code here
            pass
        except Exception as e:
            log.error(f"Exception caught: {e}")
            if log.level == logging.DEBUG:
                log.tracedebug()  # Print full traceback

    Testing all log levels::

        from megamicros import log
        
        log.setLevel('debug')
        log.debug('This is a debug message')
        log.info('This is an info message') 
        log.warning('This is a warning message')
        log.error('This is an error message')
        log.critical('This is a critical message')

Note:
    Colors are automatically disabled on Windows systems and enabled on Unix-like systems.
    The default log file './megamicros.log' is created in the current working directory.

See Also:
    formats_str(): Get available format options and conversions
    tracedebug(): Print debug trace information

Documentation:
    Full MegaMicros documentation is available at: https://readthedoc.bimea.io
"""



import logging
import traceback
import platform

DEBUG_MODE = True
DEFAULT_LOGFILE = './megamicros.log'

class MuFormatter(logging.Formatter):
	"""Logging Formatter to add colors and count warning / errors
	
	Note that the colors are not working on Windows
	"""

	if platform.system() == 'Windows':
		red = ""
		green = ""
		blue = ""
		magenta = ""
		grey = ""
		yellow = ""
		bold_red = ""
		bold_black = ""
		reset = ""
	else:	
		green = "\x1b[32;21m"
		blue = "\x1b[34;21m"
		magenta = "\x1b[35;21m"
		grey = "\x1b[38;21m"
		yellow = "\x1b[33;21m"
		red = "\x1b[31;21m"
		bold_red = "\x1b[31;1m"
		bold_black = "\x1b[30;1m"
		reset = "\x1b[0m"

	start_format = magenta + "%(asctime)s " + reset + bold_black + "[%(levelname)s]: " + reset

	FORMATS = {
		logging.DEBUG: start_format + green + "in %(name)s (%(filename)s:%(lineno)d): %(message)s" + reset,
		logging.INFO: magenta + "%(asctime)s " + reset + "[%(levelname)s]: " + blue + "%(message)s" + reset,
        logging.WARNING: start_format + yellow + "in %(name)s (%(filename)s:%(lineno)d): %(message)s" + reset,
        logging.ERROR: start_format + red + "in %(name)s (%(filename)s:%(lineno)d): %(message)s" + reset,
        logging.CRITICAL: start_format + bold_red + "in %(name)s (%(filename)s:%(lineno)d): %(message)s" + reset
    }

	def format(self, record):
		log_fmt = self.FORMATS.get( record.levelno )
		formatter = logging.Formatter( log_fmt )
		return formatter.format( record )


mulog_ch = logging.StreamHandler()
mulog_ch.setLevel( logging.DEBUG )
mulog_ch.setFormatter( MuFormatter() )

mulog_ch2 = logging.FileHandler( DEFAULT_LOGFILE, mode='a', encoding='utf-8', delay=False, errors=None)
mulog_ch2.setLevel( logging.DEBUG )
mulog_ch2.setFormatter( MuFormatter() )

log = logging.getLogger( __name__ )
log.addHandler( mulog_ch2 )
log.addHandler( mulog_ch )
log.setLevel( logging.NOTSET )

def formats_str( arg: int|str|None = None ) -> int|str|list[dict[str, str|int]]|None :
	"""Convert between log level names and logging constants.

	This function provides bidirectional conversion between string labels
	('debug', 'info', etc.) and their corresponding logging module constants.

	Args:
		arg: Can be:
			- int: logging constant (returns corresponding string label)
			- str: string label (returns corresponding logging constant) 
			- None: returns list of all available formats
			
	Returns:
		- str: label name if arg is int
		- int: logging constant if arg is str  
		- list: all available formats if arg is None
		- None: if arg type is invalid
		
	Examples:
		>>> formats_str(logging.DEBUG)
		'debug'
		>>> formats_str('info')
		20
		>>> formats_str()
		[{'label': 'debug', 'format': 10}, ...]
	"""
	formats: list[dict[str, str|int]] = [
		{'label': 'debug', 'format': logging.DEBUG },
		{'label': 'info', 'format': logging.INFO },
		{'label': 'warning', 'format': logging.WARNING },
		{'label': 'error', 'format': logging.ERROR },
		{'label': 'critical', 'format': logging.CRITICAL },
	]

	if type( arg ) == int:
		return next( ( format['label'] for format in formats if format['format']==arg ), None )
	elif type( arg ) == str:
		return next( ( format['format'] for format in formats if format['label']==arg ), None )
	elif arg is None:
		return formats
	else:
		return None
	

def tracedebug():
	"""Print traceback information if debug level is active.
    
    Only prints the full exception traceback when the current log level
    is set to DEBUG. Does nothing at other log levels.
    
    Examples:
        >>> try:
        ...     1/0
        ... except:
        ...     log.tracedebug()  # Only prints if log.level == logging.DEBUG
    """
	if log.level == logging.DEBUG:
		print( traceback.format_exc() )



# megamicros.mqtt.py mqtt client handler for MegaMicros libraries
#
# Copyright (c) 2023 Sorbonne Université
# Author: bruno.gas@sorbonne-universite.fr
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Megamicros documentation is available on https://readthedoc.bimea.io

!!! WARNING!!!
The MQTT client should be created and should connect to the broker outside the MqttHandler
In fact the client should be given as argument to the MqttHandler

!!! WARNING!!!


Release 2.0.0 of the Paho Python MQTT includes breaking changes; 
this means that code written for v1.x will not work without some (minimal) modifications. 
As v2.0.0 was only released a few days ago (11th Feb 2024) most examples, including the one you reference, will not work.
See bellow for the migration guide:
https://eclipse.dev/paho/files/paho.mqtt.python/html/migrations.html
https://github.com/eclipse/paho.mqtt.python/blob/master/docs/migrations.rst
so far the code bellow is working with the 1.6.1 version of the paho-mqtt library.
As such it breaks with the 2.0 version. In the mean time.
In the mean time we only changed the 'client = mqtt.Client(client_id)' to 'client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION1, client_id)'.
It seems enough to make the code work with the 2.0 version of the paho-mqtt library.

See: https://www.emqx.com/en/blog/how-to-use-mqtt-in-python 
"""

import paho.mqtt.client as mqtt
from .exception import MuException
from .log import log, logging

DEFAULT_BROKER_HOST = 'localhost'
DEFAULT_BROKER_PORT = 1883
DEFAULT_CLIENT_NAME = 'megamicros/unknown'
DEFAULT_TOPIC = 'megamicros/unknown/unknown/log'
DEFAULT_QOS = 1
DEFAULT_LEVEL = logging.NOTSET

MQTT_ERROR_0 = 'Connection successful'
MQTT_ERROR_1 = 'incorrect protocol version'
MQTT_ERROR_2 = 'invalid client identifier'
MQTT_ERROR_3 = 'server unavailable'
MQTT_ERROR_4 = 'bad username or password'
MQTT_ERROR_5 = 'not authorised'


class MqttClient :
    __client: mqtt.Client
    __connected: bool
    __host: str
    __port: int
    __name: str
    __onMessageCallback = None
    __onMessageCallbackUserdata = None


    def __init__( self, host=DEFAULT_BROKER_HOST, port=DEFAULT_BROKER_PORT, name=DEFAULT_CLIENT_NAME ) :
        self.__connected = False
        self.__host = host
        self.__port = port
        self.__name = name
        self.__onMessageCallback = None
        self.__onMessageCallbackUserdata = None

        def on_connect(client, userdata, flags, rc):
            if rc == 0:
                log.info( f" .Connected on MQTT broker")
            else:
                log.info( f" .Failed to connect to MQTT broker with return code: '{rc}'")
        
        try :
            self.__client = mqtt.Client( mqtt.CallbackAPIVersion.VERSION1, name )
            self.__client.user_data_set( self )
            self.__client.on_connect = on_connect
            self.__client.connect( host=host, port=port, keepalive=60, bind_address="" )
            self.__connected = True


        except Exception as e:
            log.error( f"MQTT broker connection failed: {e}" )
            raise

    def __del__( self ) :
        self.__client.disconnect()

    def is_connected( self ) -> bool : 
        return self.__connected

    def publish( self, message: str, topic: str, qos: int=1 ) :
        self.__client.publish( topic, message, qos, retain=False )

    def subscribe(self, topic: str, on_message=None, userdata=None):
        def on_message_default(client, userdata, msg):
            print(f"Received `{msg.payload.decode()}` from `{msg.topic}` topic")

        self.__onMessageCallback = on_message if on_message is not None else self.__onMessageCallback
        self.__onMessageCallbackUserdata = userdata if userdata is not None else self.__onMessageCallbackUserdata

        self.__client.subscribe(topic)
        self.__client.on_message = on_message_default if self.__onMessageCallback is None else self.__onMessageCallback
        self.__client.user_data_set( self.__onMessageCallbackUserdata )

    def onMessage(self, callback, userdata=None):
        self.__onMessageCallback = callback
        self.__onMessageCallbackUserdata = userdata
        

    def run(self, topic: str, on_message=None, userdata=None ):
        self.subscribe( topic, on_message, userdata )
        self.__client.loop_forever()


class MqttPubHandler( logging.Handler ) :

    __client: MqttClient
    __topic: str
    __qos: int

    def __init__( self, client: MqttClient, topic=DEFAULT_TOPIC, qos=DEFAULT_QOS, level=DEFAULT_LEVEL ) :
        try :
            super().__init__( level )
            if not client.is_connected():
                raise MuException( 'Cannot init logging Handler: MQTT client is not connected' )
            self.__client = client
            self.__topic = topic
            self.__qos = qos

        except Exception as e:
            log.error( f"MQTT broker connection failed: {e}" )
            raise

    def getTopic( self ) -> str :
        return self.__topic


    def __del__(self) :
        self.__client = None


    def emit( self, record ) :
        message = record.getMessage()
        self.__client.publish( message=message, topic=self.__topic, qos=self.__qos )






# Megamicros_tools.acoustics.antenna.py
#
# Copyright (c) 2024 Bimea
# Author: bruno.gas@bimea.io
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

""" Provide antenna definitions

Documentation
-------------
MegaMicros documentation is available on https://readthedoc.bimea.io
"""

import numpy as np
from scipy.spatial.distance import cdist
from megamicros.log import log
from megamicros.exception import MuException


"""
Declare the square antenna of 32 mems used for the Mosellerie and Le Prehaut POCs 

vue de dessus


                  (back)
      24  25  26  27  28  29  30  31

                    F3
     --x---x---x---x---x---x---x---x--
7    x(0) (1) (...)              (0) x    16
     |                               | 
6    x                           (1) x    17
     |                               |
5    x                         (...) x    18
     |                               |
4    x                               x    19
  F0 |                               |  F2
3    x                               x    20
     |                               |
2    x (...)                         x    21
     |                               |
1    x (1)                           x    22
     |                               |
0    x (0)              (...) (1) (0)x    23
     --x---x---x---x---x---x---x---x--
  (0,0)             F1

       15  14  13  12  11  10  9   8              
                    !
                    !
                    !
              vers la porte (front)

MEMs numbering:
F0: 0,1,2,3,4,5,6,7
F1: 8, 9, 10, 11, 12, 13, 14, 15
F2: 16, 17, 18, 19, 20, 21, 22, 23
F3: 24, 25, 26, 27, 28, 29, 30, 31
"""


Mu32_Mems32_JetsonNano_0001 = {
    "name": "Mu32_Mems32_JetsonNano_0001",
    "mems": list( np.array( [
        [0, 3.82, 0], [0, 9.82, 0], [0, 15.82, 0], [0,  21.82, 0], [0,  27.82, 0], [0, 33.82, 0], [0, 39.82, 0], [0, 45.82, 0],
        [45.81, 0, 0], [39.81, 0, 0], [33.81, 0, 0], [27.81, 0, 0], [21.81, 0, 0], [15.81, 0, 0], [9.81, 0, 0] , [3.81, 0, 0],
        [49.63, 45.82, 0], [49.63, 39.82, 0], [49.63, 33.82, 0], [49.63,  27.82, 0], [49.63,  21.82, 0], [49.63, 15.82, 0], [49.63, 9.82, 0], [49.63, 3.82, 0],
        [3.81, 49.63, 0], [9.81, 49.63, 0], [15.81, 49.63, 0], [21.81, 49.63, 0], [27.81, 49.63, 0], [33.81, 49.63, 0], [39.81, 49.63, 0], [45.81, 49.63, 0]
    ] )/100 - 0.25 ),
    "comment": "Square antenna dimensions used in Lepreau experiments",
    "unit": "m",
}


def gen_circular( radius: float, mems_number: int, center: np.ndarray=[0,0,0], plane: str="XY" ):
    """ Generate a circular antenna

    Parameters
    ----------
    radius: float
        The radius of the circular antenna
    mems_number: int
        The number of MEMs in the circular antenna
    center: np.ndarray
        The center of the circular antenna
    plane: str
        The plane of the circular antenna. Default is "XY"

    Returns
    -------
    circular_antenna: dict
        The circular antenna
    """
    if plane == "XY":
        angle = np.linspace( 0, 2*np.pi, mems_number, endpoint=False )
        circular_antenna = {
            "name": "circular_antenna",
            "mems": list( np.array( [ center[0] + radius * np.cos( angle ), center[1] + radius * np.sin( angle ), np.zeros( mems_number ) ] ).T ),
            "comment": "Circular antenna",
            "unit": "m",
        }
    else:
        raise MuException( "The plane is not yet implemented" )
    
    return circular_antenna


def gen_helicoidal( radius: float, mems_number: int, center: np.ndarray=[0,0,0], laps: float=2, plane: str="XY" ):
    """ Generate a helicoidal antenna

    Parameters
    ----------
    radius: float
        The radius of the helicoidal antenna
    mems_number: int
        The number of MEMs in the helicoidal antenna
    center: np.ndarray
        The center of the helicoidal antenna
    laps: float
        NUmber of laps to be completed. Default is 2
    plane: str
        The plane of the helicoidal antenna. Default is "XY"

    Returns
    -------
    helicoidal_antenna: dict
        The helicoidal antenna
    """
    if plane == "XY":
        angle = np.linspace( 0, 2*np.pi*laps, mems_number, endpoint=False )
        radius = np.linspace( 0, radius, mems_number, endpoint=False )
        helicoidal_antenna = {
            "name": "helicoidal_antenna",
            "mems": list( np.array( [ center[0] + radius * np.cos( angle ), center[1] + radius * np.sin( angle ), np.zeros( mems_number ) ] ).T ),
            "comment": "Helicoidal antenna",
            "unit": "m",
        }
    else:
        raise MuException( "The plane is not yet implemented" )
    
    return helicoidal_antenna




"""
antenna= {'positions': np.array(
[[-0.2261063,  -0.2217998,   0.        ],
[-0.2231343,  -0.16230868,  0.        ],
[-0.23502814,  0.01106646,  0.        ],
[-0.23505722,  0.07143718,  0.        ],
[-0.23869585,  0.13492614,  0.        ],
[-0.241065,    0.19720244,  0.        ],
[-0.20675762,  0.23860315,  0.        ],
[-0.14724884,  0.23868911,  0.        ],
[-0.08815056,  0.23800337,  0.        ],
[-0.02524838,  0.23605316,  0.        ],
[ 0.03711273,  0.23184893,  0.        ],
[ 0.09778664,  0.22929929,  0.        ],
[ 0.20982932,  0.21545461,  0.        ],
[ 0.25022585,  0.17556075,  0.        ],
[ 0.24945468,  0.11548598,  0.        ],
[ 0.2482488,   0.05719138,  0.        ],
[ 0.25352816,  0.00069583,  0.        ],
[ 0.25355045, -0.17305451,  0.        ],
[ 0.26728441, -0.23428049,  0.        ],
[ 0.23313417, -0.27617064,  0.        ],
[ 0.16906521, -0.27623144,  0.        ],
[-0.01177931, -0.2639107,   0.        ],
[-0.0767392,  -0.26609785,  0.        ],
[-0.13061677, -0.26229372,  0.        ],
[-0.18359293, -0.25536996,  0.        ]] ),
'mems':[0,1,4,5,6,7,8,9,10,11,12,13,15,16,17,18,19,22,23,24,25,28,29,30,31],
'available_mems': [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31]
}
np.save( 'Antenna-square-JetsonNano-0001.npy', antenna )
"""




import os
try:
    from importlib.metadata import version
except ImportError:
    # Fallback for Python < 3.8
    from importlib_metadata import version

def _get_version():
    try:
        # Try to read from VERSION file (for development)
        version_file = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), 'VERSION')
        if os.path.exists(version_file):
            with open(version_file, 'r') as f:
                return f.read().strip()
    except:
        pass
    
    try:
        # Try to get version from installed package (for installed package)
        return version('megamicros')
    except:
        pass
    
    # Fallback version
    return "unknown"

__version__ = _get_version()

welcome_msg = '-'*20 + '\n' + 'Megamicros python library\n \
Copyright (C) 2024-2025 Bimea\n \
This program comes with ABSOLUTELY NO WARRANTY; for details see the source code\'.\n \
This is free software, and you are welcome to redistribute it\n \
under certain conditions; see the source code for details.\n' + '-'*20 + '\n' + '\
MegaMicros documentation is available on https://readthedoc.bimea.io.\n' + '-'*20